import json
from pathlib import Path
from typing import Union
from uuid import UUID


# helper functions

def create_file(path: Path) -> None:
    path.mkdir(parents=True, exist_ok=True)


def check_path(path: Union[Path, str]) -> Path:
    if not isinstance(path, Path):
        return Path(path)
    return path


def prepare_data_for_leaf(obj: dict) -> dict:
    new = obj.copy()
    data = new.pop('data')
    new.update(data)
    return new


def open_json(path: Union[str, Path], encoding='utf8'):
    with open(path, 'r', encoding=encoding) as file:
        return json.load(file)

# helper class
class UUIDEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, UUID):
            # if the obj is uuid, we simply return the value of uuid
            return str(obj)
        return json.JSONEncoder.default(self, obj)



from typing import Type

from metagen.register import register
from metagen.base import LeafABC


def find_ureferenced_key(keyName: str)-> list[LeafABC]:
    uuids = register.table['key'].unique()

    try:
        keys = register.table[keyName].unique()
    except KeyError:
        raise KeyError(f'{keyName} did not find among the element attributes in the register')

    for key in keys:
        pass


if  __name__  == '__main__':
    from pathlib import Path
    from metagen import Generator

    path = Path('C:\\Users\\micha\\PycharmProjects\\app-esaWorldWater\\fixtures.json')

    gen = Generator()
    gen.load_fixtures(path)

    pass



import pytest
from metagen.presets import PresetBackgroundMaps
from metagen.components import MapBackgroundLayer


def test_background_use_enum():
    assert MapBackgroundLayer(**PresetBackgroundMaps.CartoDB_LightNoLabels)


from typing import Union, Dict
from uuid import UUID
from abc import ABC
from pydantic import BaseModel

from metagen.base import Leaf, BaseModelWithDynamicKey, set_key_from_input


class BaseModelArbitrary(BaseModel):
    pass

    class Config:
        arbitrary_types_allowed = True


class Component(BaseModel, ABC):
    pass


class Filter(BaseModelWithDynamicKey):
    __root__: Dict[str, UUID]

    @classmethod
    def set(cls, key_name: str, key: Union[str, UUID, Leaf]):
        uuid = set_key_from_input(key)
        return cls(**{key_name: uuid})




from pydantic import BaseModel, Field
from abc import ABC, abstractmethod, ABCMeta
from pathlib import Path
from typing import Union, List
import json

from metagen.base import Leaf
from metagen.utils import create_file, check_path, open_json, UUIDEncoder
from metagen.metadata import ElementFactory, element_factory
from metagen.register import register, Register


# serialization & deserialization
class Serializer(BaseModel, ABC):

    @abstractmethod
    def to_dict(self):
        pass

    @abstractmethod
    def to_json(self, path: Union[Path, str]) -> None:
        pass


class JSONSerializer(Serializer):
    structure: dict = Field(default={})

    def to_dict(self) -> dict:
        for element in register.get_elements():
            nodes = element.__nodes__().split('.')
            self.set_node(self.structure, nodes, element)
        return self.structure

    def set_node(self, structure: dict, nodes: list, element: Leaf):
        node = nodes.pop(0)
        if len(nodes) > 0:
            if not structure.get(node):
                structure[node] = {}
            self.set_node(structure[node], nodes, element)
        else:
            if not structure.get(node):
                structure[node] = []
            structure[node].append(element.to_dict())

    def to_json(self, path: Union[Path, str]) -> None:
        structure = self.to_dict()

        path = check_path(path)

        if not path.parent:
            create_file(path.parent)

        with open(path, 'w') as file:
            json.dump(structure, file, indent=6, cls=UUIDEncoder)


class DeSerializer(BaseModel, ABC):

    @abstractmethod
    def load(self, path: Path, **kwargs) -> None:
        pass


class JSONDeserializer(DeSerializer):
    factory: ElementFactory = Field(default=element_factory)

    def load(self, path: Path, encoding='utf8') -> None:

        path = check_path(path)
        obj = open_json(path, encoding)
        for node, structure in obj.items():
            self._parse(node, structure)

    def _parse(self, nodes: str, obj: Union[dict, list]) -> None:

        if isinstance(obj, dict):
            for node, structure in obj.items():
                self._parse(f'{nodes}.{node}', structure)
        elif isinstance(obj, list):
            for data in obj:
                self.factory.create_element(nodes, data)


# generator
class GeneratorABC(BaseModel, ABC):
    serializer: Serializer
    deserializer: DeSerializer

    @property
    @abstractmethod
    def register(self) -> Register:
        pass


class Generator(GeneratorABC):
    serializer: Serializer = Field(default=JSONSerializer())
    deserializer: DeSerializer = Field(default=JSONDeserializer())

    @property
    def register(self) -> Register:
        return register

    def load_fixtures(self, path: Path, encoding='utf8') -> None:
        self.deserializer.load(path, encoding=encoding)

    def to_dict(self) -> dict:
        return self.serializer.to_dict()

    def to_json(self, path: Union[str, Path]) -> None:
        self.serializer.to_json(path)

    def get_element_by_nameInternal(self, name: str) -> Leaf:
        """Return element of given nameInternal"""
        if self.register.get_by_name(name):
            return self.register.get_by_name(name)
        else:
            raise ValueError(f'Element with nameInternal {name} did not find')

    def get_element_by_uuid(self, uuid: str) -> Leaf:
        """Return element of given uuid"""
        if register.get_by_uuid(uuid):
            return register.get_by_uuid(uuid)
        else:
            raise ValueError(f'Element with uuid {uuid} did not find')

    def get_elements_by_type(self, element: Leaf) -> List[Leaf]:
        """Return list of all elements of given element type"""
        return [v for k, v in self.register.name.items() if isinstance(v, element.__wrapped__)]

    def get_elements_by_name(self, name: str) -> List[Leaf]:
        """Return list of elements that internal name contains part of input string"""
        return [v for k, v in self.register.name.items() if k.__contains__(name)]


from metagen.generator import Generator
from metagen.metadata import Application, Configuration, Scope, Place, Period, LayerTemplate, Attribute, SpatialWMS, \
    RelationSpatial, SpatialAttribute, SpatialWMTS, SpatialVector, Style, Case, RelationAttribute, Tag


from abc import ABC, abstractmethod
from uuid import UUID
from typing import Type, Any, List
from functools import wraps
from warnings import warn
from pandas import DataFrame
import pandas
from pydantic import BaseModel, Field
import weakref

from metagen.base import LeafABC


# TODO: Solve weak references


class Register(BaseModel, ABC):

    @abstractmethod
    def get_elements(self) -> List[Type[LeafABC]]:
        pass

    @abstractmethod
    def add(self, element: Type[LeafABC]) -> None:
        pass

    @abstractmethod
    def check_register(self, element: Type[LeafABC]) -> bool:
        pass

    @abstractmethod
    def get_by_name(self, name: str) -> Type[LeafABC]:
        pass

    @abstractmethod
    def get_by_hash(self, hash: int) -> Type[LeafABC]:
        pass

    @abstractmethod
    def get_by_uuid(self, uuid: UUID) -> Type[LeafABC]:
        pass

    class Config:
        arbitrary_types_allowed = True


# register
class DictRegister(Register):
    hashs: dict = Field(default_factory=dict)
    uuid: dict = Field(default_factory=dict)
    name: dict = Field(default_factory=dict)

    def get_elements(self) -> List[Type[LeafABC]]:
        return [element for element in self.hashs.values()]

    def add(self, element: Type[LeafABC]) -> None:
        if not self.check_register(element):
            self.hashs.update({hash(element): element})
            self.uuid.update({str(element.key): element})
            self.name.update({element.nameInternal: element})
        else:
            raise ValueError(f'PTR element "{element.__class__.__name__}" with nameInternal: {element.nameInternal}, '
                             f'key: {element.key} and hash: {hash(element)} already exist')

    def check_register(self, element: Type[LeafABC]) -> bool:
        return all([self.hashs.get(hash(element)), self.name.get(element.nameInternal)])

    def get_by_name(self, name: str) -> Type[LeafABC]:
        return self.name.get(name)

    def get_by_hash(self, hash: int) -> Type[LeafABC]:
        return self.hashs.get(hash)

    def get_by_uuid(self, uuid: str) -> Type[LeafABC]:
        if UUID(uuid):
            return self.uuid.get(uuid)


def prepare_element4pandas(element: Type[LeafABC]) -> dict:
    """Helper function, creates dict for allement to concat later the element dict into pandas dataframe"""
    element_dict = {k:v for k, v in element.dict().items() if not isinstance(v, (list, dict, tuple))}
    element_dict['element'] = weakref.ref(element)
    element_dict['type'] = element.__class__.__name__
    element_dict['hash'] = hash(element)
    return element_dict


class PandasRegister(Register):
    table: DataFrame = Field(default_factory=DataFrame)
    element_instances: list = Field(default_factory=list)

    def get_elements(self) -> List[Type[LeafABC]]:
        return [element() for element in self.table['element'].values]

    def add(self, element: Type[LeafABC]) -> None:
        """Add element to register. If element hash in register rise error"""
        self.element_instances.append(element)
        element_pd = DataFrame(prepare_element4pandas(element), index=['key'])
        self.table = pandas.concat([self.table, element_pd], ignore_index=True, axis=0)

    def check_register(self, element: Type[LeafABC])-> bool:
        try:
            self.get_by_hash(hash(element))
            return True
        except (ValueError, KeyError):
            return False

    def get_by_name(self, name: str) -> Type[LeafABC]:
        return self.get_by_attrName(name, attrName='nameInternal')

    def get_by_hash(self, hash: int) -> Type[LeafABC]:
        return self.get_by_attrName(hash, attrName='hash')

    def get_by_uuid(self, uuid: UUID) -> Type[LeafABC]:
        return self.get_by_attrName(uuid, attrName='key')

    def get_by_attrName(self, value: Any, attrName: str) -> Type[LeafABC]:
        """Find element in register based on columne name and ist value """
        filter = self.table[attrName] == value
        return self.table.loc[filter]['element'].item()()



# register = DictRegister()
register = PandasRegister()


def exist_in_register(element):
    @wraps(element)
    def checking_register(*args, **kwargs):
        instance = element(*args, **kwargs)
        if register.check_register(instance):
            registered_element = register.get_by_hash(hash(instance))
            warn(f'Element duplication: Element {instance.__class__.__name__} with parameters: '
                 f'{"; ".join([f"{k}: {v}" for k, v in kwargs.items()])} found in register. Element '
                 f'{registered_element.__repr__()} returned instead')
            return registered_element
        else:
            register.add(instance)
            return instance
    return checking_register


# general and helper function and classe

from typing import Union, Optional, Any, List
from pydantic import BaseModel, Field, root_validator, PrivateAttr
from pydantic.utils import ROOT_KEY
from abc import ABC, abstractmethod
from uuid import UUID, uuid4
import copy


# helper function
def make_hash(o):
    """ Makes a hash from a dictionary, list, tuple or set to any level, that contains
  only other hashable types (including any lists, tuples, sets, and
  dictionaries). """

    if isinstance(o, (set, tuple, list)):
        return tuple([make_hash(e) for e in o])

    elif not isinstance(o, dict):

        return hash(o)

    new_o = copy.deepcopy(o)
    for k, v in new_o.items():
        new_o[k] = make_hash(v)

    return hash(tuple(frozenset(sorted(new_o.items()))))


# helper class
class SingletonMeta(type):
    _instance = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instance:
            cls._instance[cls] = super(SingletonMeta, cls).__call__(*args, **kwargs)
        return cls._instance[cls]


class BaseModelWithDynamicKey(BaseModel):
    """
    Pydantic workaoround for custom dynamic key
    ref: https://stackoverflow.com/questions/60089947/creating-pydantic-model-schema-with-dynamic-key
    """

    def __init__(self, **data: Any) -> None:
        if self.__custom_root_type__ and data.keys() != {ROOT_KEY}:
            data = {ROOT_KEY: data}
        super().__init__(**data)


class LeafABC(BaseModel, ABC):
    key: Optional[UUID]

    @abstractmethod
    def __nodes__(self) -> str:
        pass

    @property
    @abstractmethod
    def hash_attrs(self) -> tuple:
        pass


# element
class Leaf(LeafABC):
    key: Optional[Union[UUID, str]] = Field(default_factory=uuid4)
    _input_pars: List[str] = PrivateAttr()

    def __init__(self, **data):
        super().__init__(**data)
        self._input_pars = [k for k in data.keys()]

    def to_dict(self):
        include = {k for k, v in self.__dict__.items() if k in self._input_pars or v is not None}
        exlude = {k for k in self.__dict__.keys()} - include

        data = self.dict(by_alias=True, exclude=exlude, include=include)
        key = data.pop('key')
        return {"key": str(key), "data": data}

    @root_validator(pre=True, allow_reuse=True)
    def set_key(cls, values: dict) -> dict:
        return {k: (v.key if isinstance(v, Leaf) else v) for k, v in values.items()}

    @property
    def hash_attrs(self) -> tuple:
        return tuple()

    def __hash__(self) -> int:
        return hash(tuple([self.__dict__.get(attr) if not isinstance(self.__dict__.get(attr), list)
                                   else tuple(self.__dict__.get(attr)) for attr in self.hash_attrs]))

    # def __hash__(self) -> int:
    #     return make_hash(self.to_dict())

    __slots__ = ('__weakref__',)

    # class Config:
    #     fields = {'_input_pars': {'exclude': True}}


def set_key_from_input(value: Union[str, UUID, Leaf]):
    """
    Helper method used as validator in pydantic model.
    For input string, Leaf or UUID return valid UUID
    """
    if isinstance(value, Leaf):
        return value.key
    if isinstance(value, str):
        return UUID(value)
    return value



import pytest
from uuid import uuid4

from metagen.metadata import Application, LayerTemplate
from metagen.register import PandasRegister
from test.test_components import layer_template_1


# @pytest.fixture(autouse=True)
# def layer_template_1():
#     return LayerTemplate(applicationKey='app', nameInternal=f'lt1', nameDisplay='lt1')


@pytest.fixture(autouse=True)
def layer_template_2():
    return LayerTemplate(applicationKey='app', nameInternal=f'lt2', nameDisplay='lt2')


@pytest.fixture(autouse=True)
def pandas_register(layer_template_1, layer_template_2):
    register = PandasRegister()
    register.add(layer_template_1)
    register.add(layer_template_2)
    return register


def test_pandas_register_add(pandas_register, layer_template_2):
    assert pandas_register.table['element'][1]() == layer_template_2


def test_pandas_register_get_by_uuid(pandas_register, layer_template_2):
    assert pandas_register.get_by_uuid(layer_template_2.key) == layer_template_2


def test_pandas_register_get_by_hash(pandas_register, layer_template_1):
    assert pandas_register.get_by_hash(hash(layer_template_1)) == layer_template_1


def test_pandas_register_get_by_name(pandas_register, layer_template_2):
    assert pandas_register.get_by_name(layer_template_2.nameInternal) == layer_template_2


def test_pandas_register_get_by_name(pandas_register, layer_template_2):
    assert pandas_register.check_register(layer_template_2) is True
    assert pandas_register.check_register(uuid4()) is False


import pytest

from metagen.metadata import Application, LayerTemplate

def test_element_duplicity():
    lt1 = LayerTemplate(applicationKey='app', nameInternal=f'lt', nameDisplay='lt')
    lt2 = LayerTemplate(applicationKey='app', nameInternal=f'lt', nameDisplay='lt')
    assert lt1.key == lt2.key



def test_stac_attribute_in_leaf():
    app = Application(name='test', nameInternal='test')
    assert hasattr(app, '_input_pars')
    assert app._input_pars == ['name', 'nameInternal']


def test_exlude_stac_attribute_from_leaf():
    app = Application(name='test', nameInternal='test')
    dict = app.dict()
    assert dict.get('_input_pars') is None


def test_to_dict():
    """testing of exluding atributes with none values except those specified as none during the inicialization"""
    lt = LayerTemplate(applicationKey=None, nameInternal='test')
    lt_dict = lt.to_dict()
    data = lt_dict.get('data')
    assert all([k in ['applicationKey', 'nameInternal'] for k in data.keys()])
    assert data.get('applicationKey') is None